/* BnuView v1.2
 * Written by Prez@lfx.org
 * 
 * For more information, see README
 *
 */

#include <vga.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <font.h>
#include <widget.h>

#define w 16
#define h 16

#define BPP1	1
#define GB	2
#define NES	3

#define _8X8	1
#define _8X16	2
#define _16X16	3

#define WHITE	15
#define BLACK	0
#define LTGRAY	7
#define DRGRAY	8

int fp;

int bpp1_tile(int loc, int x, int y) {
	int r,c;
	char data[8];

	if (!fp) return(-1);

	lseek(fp, loc, SEEK_SET);
	read(fp, data, 8);


	for ( r = 0 ; r < 8 ; r++ ) {
		for ( c = 0 ; c < 8 ; c++ ) {
			if ( data[r] & 1<<c ) vga_setcolor(WHITE);
			else vga_setcolor(BLACK);
			vga_drawpixel(x+(7-c),y+r);
		}
	}

	return(0);
}

int gb_tile(int loc, int x, int y) {
	int r,c;
	char data[16];

	if (!fp) return(-1);

	lseek(fp, loc, SEEK_SET);
	read(fp, data, 16);

	for ( r = 0 ; r < 16 ; r+=2 ) {
		for ( c = 0 ; c < 8 ; c++ ) {
			if ( data[r] & 1<<c && data[r+1] & 1<<c ) vga_setcolor(WHITE);
			else if ( data[r] & 1<<c && ! (data[r+1] & 1<<c) ) vga_setcolor(LTGRAY);
			else if ( ! (data[r] & 1<<c) && data[r+1] & 1<<c ) vga_setcolor(DRGRAY);
			else vga_setcolor(BLACK);

			vga_drawpixel(x+(7-c),y+(r/2));
		}
	}

	return(0);
}

int nes_tile(int loc, int x, int y) {
	int r,c;
	char data[16];

	if (!fp) return(-1);

	lseek(fp, loc, SEEK_SET);
	read(fp, data, 16);

	for ( r = 0 ; r < 8 ; r++ ) {
		for ( c = 0 ; c < 8 ; c++ ) {
			if ( data[r] & 1<<c && data[r+8] & 1<<c ) vga_setcolor(WHITE);
			else if ( data[r] & 1<<c && ! (data[r+8] & 1<<c) ) vga_setcolor(LTGRAY);
			else if ( ! (data[r] & 1<<c) && data[r+8] & 1<<c ) vga_setcolor(DRGRAY);
			else vga_setcolor(BLACK);

			vga_drawpixel(x+(7-c),y+r);
		}
	}

	return(0);
}

int draw_file(int loc, int mode, int bpp, int x, int y) {
	int r,c;

	for ( r = 0 ; r < h ; r++ ) {
		for ( c = 0 ; c < w ; c++ ) {
		  if ( mode == BPP1 ) bpp1_tile(loc+((c+(r*w))*(8*bpp)), x+(c*8), y+(r*8));
		  else if ( mode == GB ) gb_tile(loc+((c+(r*w))*(8*bpp)), x+(c*8), y+(r*8));
		  else if ( mode == NES ) nes_tile(loc+((c+(r*w))*(8*bpp)), x+(c*8), y+(r*8));
		}
	}
}

int draw_file_8x16(int loc, int mode, int bpp, int x, int y) {
	int r,c;

	for ( r = 0 ; r < h/2 ; r++ ) {
		for ( c = 0 ; c < w*2 ; c+=2 ) {

		if ( mode == BPP1 ) {
			bpp1_tile(loc+((c+(r*w))*(8*bpp)), x+((c/2)*8), y+((r*2)*8));
			bpp1_tile((loc+((c+(r*w))*(8*bpp))+(8*bpp)), x+((c/2)*8), y+(((r*2)*8)+8));
		}

		if ( mode == GB ) {
			gb_tile(loc+((c+(r*w))*(8*bpp)), x+((c/2)*8), y+((r*2)*8));
			gb_tile((loc+((c+(r*w))*(8*bpp))+(8*bpp)), x+((c/2)*8), y+(((r*2)*8)+8));
		}

		if ( mode == NES ) {
			nes_tile(loc+((c+(r*w))*(8*bpp)), x+((c/2)*8), y+((r*2)*8));
			nes_tile((loc+((c+(r*w))*(8*bpp))+(8*bpp)), x+((c/2)*8), y+(((r*2)*8)+8));
		}

		}
	}
}

int main(int argc, char ** argv) {
	int loc = 0;
	int input;
	int mode = BPP1;
	int bpp = 1;
	int form = _8X8;
	int eof;
	int update = 1;
	int adupdate = 1;
	int alupdate = 1;
	char addr[40];
	char datain[0xFF];

	if ( argc < 2 ) exit(printf("Usage: %s <file>\n", argv[0]));

	fp = open(argv[1],O_RDONLY);
	if ( fp < 0 ) exit(printf("Error loading file '%s'..\n", argv[1]));

	eof = lseek(fp, 0, SEEK_END);

	vga_init();
	vga_setmode(G320x240x256);

	vga_clear();

	while (1) {

		if ( alupdate == 1 ) {
			draw_window(argv[1],0,0,(8*16)+6,(8*16)+2);

			draw_string("BnuView v1.2 Written by Prez@lfx.org",320-get_vwf("BnuView v1.2 Written by Prez@lfx.org"),240-16, WHITE,1);
			draw_string("Controls:\nw  Line up\nz  Line down\ns  Byte up\na  Byte down\ne  Page up\nx  Page down\n\nj  Jump to..", (9*16), 0, WHITE,1);
			draw_string("1  BPP1\n2  Gameboy\n3  NES\n\n8  8x8 tiles\n9  8x16 tiles\n\nq  Quit", (8*30), 16, WHITE,1);

			sprintf(addr,"End: 0x%08X", eof);
			draw_string(addr,0,((8*16)+(30))+(1*16), WHITE, 1);

			alupdate = 0;
		}

		if ( adupdate == 1 ) {
			/* Draw the current location in the file to screen */
			if ( form == _8X8 ) draw_file(loc, mode, bpp, 3, 23+3);
			else if ( form == _8X16 ) draw_file_8x16(loc, mode, bpp, 3, 26);

			/* Update any dynamic screen information */
			sprintf(addr,"Offset: 0x%08X  ", loc);
			draw_string(addr,0,((8*16)+(30))+(0*16), WHITE, 1);

			adupdate = 0;
		}

		if ( update == 1 ) {
			if ( mode == BPP1 ) draw_string("Bpp1   ",0,((8*16)+(30))+(2*16), WHITE, 1);
			else if ( mode == GB ) draw_string("Gameboy ",0,((8*16)+(30))+(2*16), WHITE, 1);
			else if ( mode == NES ) draw_string("NES     ",0,((8*16)+(30))+(2*16), WHITE, 1);


			if ( form == _8X8 ) draw_string("8x8 tile mode ",0,((8*16)+(30))+(3*16), WHITE, 1);
			else if ( form == _8X16 ) draw_string("8x16 tile mode ",0,((8*16)+(30))+(3*16), WHITE, 1);

			update = 0;
		}

		/* Wait for some key input, than what it is */
		input = vga_getch();
		if ( input == 'w' ) loc-=((bpp*8)*w), adupdate = 1;
		else if ( input == 'z' ) loc+=((bpp*8)*w), adupdate = 1;
		else if ( input == 'a' ) loc--, adupdate = 1;
		else if ( input == 's' ) loc++, adupdate = 1;
		else if ( input == 'e' ) loc-=((bpp*8)*w)*h, adupdate = 1;
		else if ( input == 'x' ) loc+=((bpp*8)*w)*h, adupdate = 1;

		else if ( input == '1' ) mode = BPP1, bpp = 1, update = 1, adupdate = 1;
		else if ( input == '2' ) mode = GB, bpp = 2, update = 1, adupdate = 1;
		else if ( input == '3' ) mode = NES, bpp = 2, update = 1, adupdate = 1;

		else if ( input == '8' ) form = _8X8, update = 1, adupdate = 1;
		else if ( input == '9' ) form = _8X16, update = 1, adupdate = 1;

		else if ( input == 'j' ) {
			update = 1, adupdate = 1, alupdate = 1;
			draw_get("jump to..",datain,10,10,10);
			sscanf(datain,"%x",&loc);
			vga_clear();
		}

		else if ( input == 'q' ) {
			vga_setmode(TEXT);
			exit();
		}


		if ( loc > eof - ((bpp*8)*16*16) ) loc = eof - ((bpp*8)*16*16);
		if ( loc < 0 ) loc = 0;

	}

	/*we should never get here.. but incase we do..*/
	vga_getch();
	vga_setmode(TEXT);
	return(0);
}		
